import logging
import numpy as np
from tqdm import tqdm
import torch
from graphmae.datasets_bulid import Gbuild_z
import warnings
from scipy.stats import pearsonr
from sklearn.metrics import mean_squared_error
from sklearn.preprocessing import StandardScaler
warnings.filterwarnings("ignore", category=UserWarning)
from graphmae.utils import (
    build_args,
    create_optimizer,
)
from sklearn.mixture import GaussianMixture
import random
import numpy as np
import pandas as pd
from graphmae.utils import build_args

import torch.nn as nn
import dgl
random_seed =4

random.seed(random_seed) 
np.random.seed(random_seed )
torch.manual_seed(random_seed )
torch.cuda.manual_seed_all(random_seed )
torch.backends.cudnn.deterministic = True

import dgl
dgl.random.seed(random_seed)

import matplotlib.pyplot as plt
from graphmae.models import IDEC_process

logging.basicConfig(format="%(asctime)s - %(levelname)s - %(message)s", level=logging.INFO)
import math

def earlyStop(loss_list, epoch,max_epoch,delta_label):

    if (epoch == max_epoch - 1):
        return True

    elif (epoch > 10) and (abs(loss_list[epoch] - loss_list[epoch - 1])< 1e-4) and (
            abs(loss_list[epoch - 1] -loss_list[epoch - 2])< 1e-4) and (
            abs(loss_list[epoch - 2] -loss_list[epoch - 3])< 1e-4) and (
            abs(loss_list[epoch - 3] -loss_list[epoch - 4])< 1e-4):
        return True

    elif (epoch > 10) and (loss_list[epoch] > loss_list[epoch - 1]) and (
            loss_list[epoch - 1] > loss_list[epoch - 2]) and (
                   loss_list[epoch-2] >  loss_list[epoch - 3])and (
                   loss_list[epoch-3] >  loss_list[epoch - 4]):
        return True
    elif (epoch > 10) and (delta_label<0.001):
        return True
    else:
        return False

def pretrain_ae(model,graph,optim_type,weight_decay,lr,use_scheduler):
    '''
    pretrain autoencoder
    '''
    optimizer_l = create_optimizer(optim_type, model, lr, weight_decay)
    max_epoch =50
    if use_scheduler:
        logging.info("Use schedular")
        scheduler = lambda epoch: (1 + np.cos((epoch) * np.pi / max_epoch)) * 0.5

        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer_l, lr_lambda=scheduler)
    else:
        scheduler = None



    epoch_iter = tqdm(range(max_epoch))

    for epoch in epoch_iter:
        print(epoch)
        model.train()
        loss_list = []

        x_bar, q = model(graph, graph.ndata["feat"])
        x_bar = x_bar.to(torch.float64)
        loss = F.mse_loss(x_bar, graph.ndata["feat"])
        optimizer_l.zero_grad()

        loss.backward()
        optimizer_l.step()
        loss_list.append(loss.item())

        if scheduler is not None:
            scheduler.step()
    return model

def target_distribution(q):
    weight = q**2 / q.sum(0)
    return (weight.t() / weight.sum(1)).t()
from sklearn.cluster import KMeans
import torch.nn.functional as F


def train_idec(model,graph, optimizer, max_epoch,n_cluster, scheduler,gamma):
    model.eval()
    x_bar, hidden = model.ae(graph, graph.ndata["feat"])
    kmeans = KMeans(n_clusters=n_cluster, random_state=4, n_init=1, init='k-means++')
    y = kmeans.fit_predict(hidden.data.cpu().numpy())

    model.cluster_layer.data = torch.tensor(kmeans.cluster_centers_)
    y_pred_last = y 

    epoch_iter = tqdm(range(max_epoch))
    trian_loss_z = []
    for epoch in epoch_iter:
        loss_list = []

        model.eval()
        _, tmp_q = model(graph, graph.ndata["feat"])

        tmp_q = tmp_q.data
        p = target_distribution(tmp_q)

        y_pred = tmp_q.cpu().numpy().argmax(1)
        delta_label = np.sum(y_pred != y_pred_last).astype(
            np.float32) / y_pred.shape[0]
        print(delta_label)

        model.train()
        x_bar, q = model(graph, graph.ndata["feat"])
        x_bar = x_bar.to(torch.float64)

        reconstr_loss =  F.mse_loss(x_bar,graph.ndata["feat"])
        kl_loss = F.kl_div(q.log(), p)
        loss =gamma * kl_loss + (1-gamma) *reconstr_loss

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if scheduler is not None:
            scheduler.step()


        loss_1=loss.item()
        loss_2= float("{:.8f}".format(loss_1))
        trian_loss_z.append(loss_2)

        if earlyStop(trian_loss_z, epoch,max_epoch,delta_label):
            return model

            break

def gcae(args,graph):

    device ="cpu"
    max_epoch = args.max_epoch


    optim_type = args.optimizer
    weight_decay=args.weight_decay
    lr = args.lr
    use_scheduler = args.scheduler
    n_clusters= args.n_cluster
    gamma = args.gamma



    model_l = IDEC_process(args)
    model_l.to(device)
    model = pretrain_ae(model_l,graph,optim_type, weight_decay,lr, use_scheduler)


    optimizer = create_optimizer(optim_type, model, lr, weight_decay)

    if use_scheduler:
        logging.info("Use schedular")
        scheduler = lambda epoch: (1 + np.cos((epoch) * np.pi / max_epoch)) * 0.5

        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=scheduler)
    else:
        scheduler = None

    model.eval()
    model_z=train_idec(model,graph, optimizer, max_epoch, n_clusters, scheduler,gamma)

    return model_z
def gcae_imputation(model,args,graph):
    torch.save(model.state_dict(),"imp_model.pt")

    model_z = IDEC_process(args)

    model_z.load_state_dict(
        torch.load("imp_model.pt"))
    model_z.eval()
    x1, _ = model_z(graph, graph.ndata["feat"])
    data_z = x1.detach().numpy()
    return data_z

if __name__ == "__main__":
    args = build_args()
    data = pd.read_csv("GSE81861_feature_processing.csv", index_col=0)
    graph, num_features = Gbuild_z.data_processing(data)
    args.num_features = num_features
    model = gcae(args, graph)  
    res = gcae_imputation(model, args, graph)
    res = pd.DataFrame(res, index=data.index, columns=data.columns)























































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































